{-# OPTIONS_GHC -fglasgow-exts #-}
module ZipperBTree where

import Control.Monad.State
import Control.Arrow ((&&&), (>>>))
import Data.List hiding (delete)

import Zipper

data BTree a = Leaf a | Branch [BTree a] deriving (Show, Eq)

data Cxt a = Top | Child { parent  :: Cxt a,   -- parent's context
                           lefts :: [BTree a], -- siblings to the left
                           rights :: [BTree a] -- siblings to the right
                         }
           deriving (Show, Eq)

type BTreeLoc    a = Loc (Cxt a) (BTree a)
type TravelBTree a = Travel (BTreeLoc a) (BTree a)

-- Utility Functions
--

-- repeat an action until the predicate becomes false
while :: Monad m => m Bool -> m a -> m [a]
while p act = do
  b <- p
  if b then liftM2 (:) act (while p act) else return []

-- Moving Around
--

-- move down to the nth child
down :: Int -> TravelBTree a
down n = modify down' >> liftM struct get where
    down' (Loc (Leaf   _ ) _) = error "Cannot go down from a leaf"
    down' (Loc (Branch []) _) = error "Cannot go down from an empty branch"
    down' (Loc (Branch cs) c) | n >= length cs = 
                                error $ "There aren't that many branches"
                              | otherwise      = 
                                let c' = Child { parent = c,
                                                 lefts  = take n cs,
                                                 rights = drop (n+1) cs }
                                in Loc { struct = (cs !! n), cxt = c' }

-- move up
up :: TravelBTree a
up = modify up' >> liftM struct get where
    up' (Loc _ Top            ) = error "Cannot go up from the top node"
    up' (Loc t (Child c ls rs)) = let t' = Branch $ ls ++ [t] ++ rs
                                  in Loc { struct = t', cxt = c }

-- move left a sibling
left :: TravelBTree a
left = modify left' >> liftM struct get where
    left' l@(Loc t (Child c ls rs)) | isFirst l = 
                                      error $ "Cannot move left from the " ++
                                              "first node"
                                    | otherwise = 
                                      let c' = Child { parent = c,
                                                       lefts  = init ls,
                                                       rights = t : rs }
                                      in Loc { struct = last ls, cxt = c' }

-- move right a sibling
right :: TravelBTree a
right = modify right' >> liftM struct get where
    right' l@(Loc t (Child c ls rs)) | isLast l  = 
                                       error $ "Cannot move right from the" ++
                                               " last node"
                                     | otherwise = 
                                       let c' = Child { parent = c,
                                                        lefts  = ls ++ [t],
                                                        rights = tail rs }
                                       in Loc { struct = head rs, cxt = c' }

-- move to the top node
top :: TravelBTree a
top = while (liftM isChild get) up >> liftM struct get

-- get the Loc corresponding to the top of the tree
-- useful for when calling traverse.
-- e.g. (getTop t) `traverse` myPath
getTop :: BTree a -> BTreeLoc a
getTop t = (Loc t Top)

-- Node classification
--

-- is the top node
isTop :: BTreeLoc a -> Bool
isTop loc = case loc of
              (Loc _ Top) -> True
              (Loc _ _  ) -> False

-- is not the top node (i.e. the child of some other node)
isChild :: BTreeLoc a -> Bool
isChild = not . isTop

-- is the first node in its siblings list?
isFirst :: BTreeLoc a -> Bool
isFirst loc = case loc of
                (Loc _ Top           ) -> True
                (Loc _ (Child _ [] _)) -> True
                (Loc _ _             ) -> False

-- is the last node in its siblings list?
isLast :: BTreeLoc a -> Bool
isLast loc = case loc of
               (Loc _ Top           ) -> True
               (Loc _ (Child _ _ [])) -> True
               (Loc _ _             ) -> False

-- BTree-specific Mutation
-- 

-- insert a subtree to the left of the current node
insertLeft :: BTree a -> TravelBTree a
insertLeft t' = modify insertLeft' >> liftM struct get where
    insertLeft' (Loc _ Top) = error "Cannot insert left of the top node"
    insertLeft' (Loc t c  ) = let c' = Child { parent = parent c,
                                               rights = t : rights c,
                                               lefts  = lefts c }
                              in Loc { struct = t', cxt = c' }

-- insert a subtree to the right of the current node
insertRight :: BTree a -> TravelBTree a
insertRight t' = modify insertRight' >> liftM struct get where
    insertRight' (Loc _ Top) = error "Cannot insert right of the top node"
    insertRight' (Loc t c  ) = let c' = Child { parent = parent c,
                                                rights = rights c,
                                                lefts  = lefts c ++ [t] }
                               in Loc { struct = t', cxt = c' }

-- insert a subtree as the last child of the current node
insertDown :: BTree a -> TravelBTree a
insertDown t' = modify insertDown' >> liftM struct get where
    insertDown' (Loc (Leaf _)    _) = error "Cannot insert down of a leaf"
    insertDown' (Loc (Branch cs) c) = let c' = Child { parent = c,
                                                       rights = [],
                                                       lefts  = cs }
                                      in Loc { struct = t', cxt = c' }

-- insert a subtree as the nth child of the current node
insertDownAt :: BTree a -> Int -> TravelBTree a
insertDownAt t' n = modify insertDA' >> liftM struct get where
    insertDA' (Loc (Leaf _)    _) = error "Cannot insert down of a leaf"
    insertDA' (Loc (Branch cs) c) = let c' = Child { parent = c,
                                                     lefts  = take (n+1) cs,
                                                     rights = drop (n+1) cs }
                                    in Loc { struct = t', cxt = c' }

-- delete the current subtree. move right if possible, otherwise left if 
-- possible, otherwise fail
delete :: TravelBTree a
delete = modify del' >> liftM struct get where
    del' (Loc _ Top) = error "cannot delete the top node"
                     -- if no siblings, move up
    del' l@(Loc t c) | isLast l && isFirst l = 
                       let c' = Child { parent = parent $ parent c,
                                        lefts  = lefts  $ parent c,
                                        rights = rights $ parent c }
                       in Loc { struct = Branch [], cxt = c' }
                     -- if the last node, move left                        
                     | isLast l  = 
                       let c' = Child { parent = parent c,
                                        lefts  = init $ lefts c,
                                        rights = rights c }
                       in Loc { struct = last $ lefts c, cxt = c' }
                     -- otherwise, just move right
                     | otherwise = 
                       let c' = Child { parent = parent c,
                                        lefts  = lefts c,
                                        rights = tail $ rights c }
                       in Loc { struct = head $ rights c, cxt = c' }