module ZipperTree where

import Control.Monad.State
import Control.Arrow (first, second)

import Zipper

data Tree a = Leaf a | Branch (Tree a) (Tree a) deriving (Show, Eq)

data Cxt a = Top
           | L (Cxt  a) (Tree a)
           | R (Tree a) (Cxt  a)
             deriving (Show)

type TreeLoc    a = Loc (Cxt a) (Tree a)
type TravelTree a = Travel (TreeLoc a) (Tree a)

-- Utility Functions
--

-- repeat an action until the predicate becomes false
while :: Monad m => m Bool -> m a -> m [a]
while p act = do
  b <- p
  if b then liftM2 (:) act (while p act) else return []

-- Movement around the tree
--

-- swap branches
swap :: TravelTree a
swap = modify left' >> liftM struct get where
    left' (Loc t (R l c)) = Loc { struct = l,
                                  cxt    = L c t }
    left' (Loc t (L c r)) = Loc { struct = r,
                                  cxt    = R t c }

-- move down a level, through the left branch
left :: TravelTree a
left = modify left' >> liftM struct get where
    left' (Loc (Leaf _    ) _) = error "Down from leaf"
    left' (Loc (Branch l r) c) = Loc { struct = l,
                                       cxt    = L c r }

-- move down a level, through the left branch
right :: TravelTree a
right = modify right' >> liftM struct get where
    right' (Loc (Leaf _    ) _) = error "Down from leaf"
    right' (Loc (Branch l r) c) = Loc { struct = r,
                                        cxt    = R l c }

-- move to a node's parent
up :: TravelTree a
up = modify up' >> liftM struct get where 
    up' (Loc _     Top) = error "Up from top"
    up' (Loc t (L c r)) = Loc { struct = Branch t r, cxt = c }
    up' (Loc t (R l c)) = Loc { struct = Branch l t, cxt = c }

-- move to the top node
top :: TravelTree a
top = while (liftM isChild get) up >> liftM struct get

-- get the Loc corresponding to the top of the tree
-- useful for when calling traverse.
-- e.g. (getTop t) `traverse` myPath
getTop :: Tree a -> TreeLoc a
getTop t = (Loc t Top)

-- Node classification
--

-- is the top node
isTop :: TreeLoc a -> Bool
isTop loc = case loc of
              (Loc _ Top) -> True
              (Loc _ _  ) -> False

-- is not the top node (i.e. the child of some other node)
isChild :: TreeLoc a -> Bool
isChild = not . isTop

-- is a left branch
isLeft :: TreeLoc a -> Bool
isLeft loc = case loc of
               (Loc _ Top    ) -> True
               (Loc _ (L _ _)) -> True
               (Loc _ (R _ _)) -> False

-- is a right branch
isRight :: TreeLoc a -> Bool
isRight loc = isTop loc || (not $ isLeft loc)